import pytest
import torch
from torch.utils.data import Dataset

from torchcross.data import (
    TaskSource,
    BatchedTaskSource,
    RandomChainTaskSource,
    ConcatTaskSource,
    TaskTarget,
)
from torchcross.data import WrapTaskSource


class DummyDataset(Dataset):
    data = torch.rand(10)
    labels = torch.randint(0, 2, (10,))

    def __getitem__(self, index):
        return self.data[index], self.labels[index]

    def __len__(self):
        return len(self.data)


@pytest.fixture
def dummy_dataset():
    return DummyDataset()


class DummyTaskSource(TaskSource):
    data = torch.rand(10)
    labels = torch.randint(0, 2, (10,))
    classes = {0: "a", 1: "b"}
    task_target = TaskTarget.MULTICLASS_CLASSIFICATION
    task_identifier = "test"

    def __getitem__(self, index):
        return self.data[index], self.labels[index]

    def __len__(self):
        return len(self.data)


@pytest.fixture
def dummy_task_source():
    return DummyTaskSource()


class TestWrapTaskSource:
    @pytest.mark.parametrize(
        "task_target",
        [TaskTarget.MULTICLASS_CLASSIFICATION, TaskTarget.BINARY_CLASSIFICATION],
    )
    def test_wrap_task_source_correctly_wraps_given_dataset(
        self, dummy_dataset, task_target
    ):
        classes = {0: "a", 1: "b"}
        wrap_task_source = WrapTaskSource(dummy_dataset, task_target, classes)

        assert wrap_task_source.labels.tolist() == dummy_dataset.labels.tolist()
        assert wrap_task_source.classes == classes
        assert wrap_task_source.task_target == task_target

        for i in range(len(dummy_dataset)):
            assert wrap_task_source[i] == dummy_dataset[i]


class TestBatchedTaskSource:
    @pytest.mark.parametrize("batch_size", [2, 3, 4, 5])
    @pytest.mark.parametrize("shuffle", [True, False])
    def test_batched_task_source_returns_correct_batches(
        self, dummy_dataset, batch_size, shuffle
    ):
        task_source = WrapTaskSource(
            dummy_dataset, TaskTarget.MULTICLASS_CLASSIFICATION, {0: "a", 1: "b"}
        )
        batched_task_source = BatchedTaskSource(task_source, batch_size, shuffle)

        for batch in batched_task_source:
            assert (
                len(batch) == batch_size
                or len(batch) == len(dummy_dataset) % batch_size
            )


class TestRandomChainTaskSource:
    def test_random_chain_task_source_iterates_over_all_batches(self, dummy_dataset):
        task_source = WrapTaskSource(
            dummy_dataset, TaskTarget.MULTICLASS_CLASSIFICATION, {0: "a", 1: "b"}
        )
        batched_task_source1 = BatchedTaskSource(task_source, 2, True)
        batched_task_source2 = BatchedTaskSource(task_source, 3, True)
        random_chain_task_source = RandomChainTaskSource(
            [batched_task_source1, batched_task_source2]
        )

        assert len(random_chain_task_source) == len(batched_task_source1) + len(
            batched_task_source2
        )


class TestConcatTaskSource:
    def test_concat_task_source_returns_correct_item_and_description(
        self, dummy_dataset
    ):
        task_source1 = WrapTaskSource(
            dummy_dataset, TaskTarget.MULTICLASS_CLASSIFICATION, {0: "a", 1: "b"}
        )
        task_source2 = WrapTaskSource(
            dummy_dataset, TaskTarget.BINARY_CLASSIFICATION, {0: "a", 1: "b"}
        )
        concat_task_source = ConcatTaskSource([task_source1, task_source2])

        for i in range(len(concat_task_source)):
            item, description = concat_task_source[i]
            if i < len(task_source1):
                assert item == task_source1[i]
                assert description.task_target == task_source1.task_target
                assert description.classes == task_source1.classes
            else:
                assert item == task_source2[i - len(task_source1)]
                assert description.task_target == task_source2.task_target
                assert description.classes == task_source2.classes



from .cross_domain_classifier import SimpleCrossDomainClassifier
from .maml import MAML, CrossDomainMAML
from .protonet import ProtoNet, SimpleCrossDomainProtoNet
from .simple import SimpleClassifier



import pytest
import torch
from torchmetrics.classification import (
    MulticlassAUROC,
    MultilabelAUROC,
    BinaryAUROC,
    MulticlassAccuracy,
    MultilabelAccuracy,
    BinaryAccuracy,
)

from torchcross.data.task import TaskTarget, TaskDescription
from torchcross.cd.metrics import (
    IndexedMulticlassAccuracy,
    Accuracy,
    IndexedMulticlassAUROC,
    AUROC,
)

# Sample task descritptions for testing
classes_multiclass = {0: "A", 1: "B", 2: "C"}
classes_binary = {0: "Negative", 1: "Positive"}
classes_multilabel = {0: "X", 1: "Y", 2: "Z"}
classes_ordinal = {1: "Low", 5: "Medium", 8: "High"}
multiclass_description = TaskDescription(
    TaskTarget.MULTICLASS_CLASSIFICATION,
    classes_multiclass,
    "multiclass_task",
    "multiclass_domain",
)
binary_description = TaskDescription(
    TaskTarget.BINARY_CLASSIFICATION,
    classes_binary,
    "binary_task",
    "binary_domain",
)
multilabel_description = TaskDescription(
    TaskTarget.MULTILABEL_CLASSIFICATION,
    classes_multilabel,
    "multilabel_task",
    "multilabel_domain",
)
ordinal_description = TaskDescription(
    TaskTarget.ORDINAL_REGRESSION,
    classes_ordinal,
    "ordinal_task",
    "ordinal_domain",
)


# Mock tensors for testing
multiclass_preds = torch.randn(10, len(classes_multiclass))
multiclass_targets = torch.randint(0, len(classes_multiclass), (10,))

binary_preds = torch.randn(10, len(classes_binary))
binary_targets = torch.randint(0, len(classes_binary), (10,))

multilabel_preds = torch.randn(10, len(classes_multilabel))
multilabel_targets = torch.randint(0, len(classes_multilabel), (10,))

ordinal_preds = torch.randn(10, len(classes_ordinal))
ordinal_target_indices = torch.randint(0, len(classes_ordinal), (10,))
index_to_ordinal_class = {i: c for i, c in enumerate(classes_ordinal)}
ordinal_targets = torch.tensor(
    [index_to_ordinal_class[t.item()] for t in ordinal_target_indices]
)


class TestIndexedMulticlassAccuracy:
    @pytest.fixture
    def metric(self):
        # Create a metric instance
        return IndexedMulticlassAccuracy(classes_ordinal)

    @pytest.fixture
    def base_metric(self):
        # Create a base metric instance
        return MulticlassAccuracy(len(classes_ordinal), average="micro")

    def test_init(self):
        # Test that the metric initializes correctly
        metric = IndexedMulticlassAccuracy(classes_multiclass)
        assert metric.num_classes == len(classes_multiclass)

    def test_forward(self, metric, base_metric):
        # Test that the metric computes the correct value
        m_result = metric(ordinal_preds, ordinal_targets)
        bm_result = base_metric(ordinal_preds, ordinal_target_indices)
        assert m_result == bm_result
        assert metric.compute() == base_metric.compute()


class TestAccuracy:
    @pytest.mark.parametrize(
        "task_description,expected",
        [
            (multiclass_description, MulticlassAccuracy),
            (binary_description, BinaryAccuracy),
            (multilabel_description, MultilabelAccuracy),
            (ordinal_description, IndexedMulticlassAccuracy),
        ],
    )
    def test_init(self, task_description, expected):
        # Test that the metric initializes correctly
        metric = Accuracy(task_description)
        assert isinstance(metric, expected)
        if expected == MulticlassAccuracy or expected == IndexedMulticlassAccuracy:
            assert metric.num_classes == len(task_description.classes)
        elif expected == MultilabelAccuracy:
            assert metric.num_labels == len(task_description.classes)
        elif expected == BinaryAccuracy:
            with pytest.raises(AttributeError):
                metric.num_classes
        else:
            pytest.fail(f"Unexpected metric type: {expected}")


class TestIndexedMulticlassAUROC:
    @pytest.fixture
    def metric(self):
        # Create a metric instance
        return IndexedMulticlassAUROC(classes_ordinal)

    @pytest.fixture
    def base_metric(self):
        # Create a base metric instance
        return MulticlassAUROC(len(classes_ordinal))

    def test_init(self):
        # Test that the metric initializes correctly
        metric = IndexedMulticlassAUROC(classes_multiclass)
        assert metric.num_classes == len(classes_multiclass)

    def test_forward(self, metric, base_metric):
        # Test that the metric computes the correct value
        m_result = metric(ordinal_preds, ordinal_targets)
        bm_result = base_metric(ordinal_preds, ordinal_target_indices)
        assert m_result == bm_result
        assert metric.compute() == base_metric.compute()


class TestAUROC:
    @pytest.mark.parametrize(
        "task_description,expected",
        [
            (multiclass_description, MulticlassAUROC),
            (binary_description, BinaryAUROC),
            (multilabel_description, MultilabelAUROC),
            (ordinal_description, IndexedMulticlassAUROC),
        ],
    )
    def test_init(self, task_description, expected):
        # Test that the metric initializes correctly
        metric = AUROC(task_description)
        assert isinstance(metric, expected)
        if expected == MulticlassAUROC or expected == IndexedMulticlassAUROC:
            assert metric.num_classes == len(task_description.classes)
        elif expected == MultilabelAUROC:
            assert metric.num_labels == len(task_description.classes)
        elif expected == BinaryAUROC:
            with pytest.raises(AttributeError):
                metric.num_classes
        else:
            pytest.fail(f"Unexpected metric type: {expected}")



from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from typing import Any

from torch.utils.data import Dataset, IterableDataset, ChainDataset

from ..task_source import TaskSource
from ..task import Task


class MetaDataset(Dataset[Task], ABC):
    task_source: TaskSource

    @abstractmethod
    def __getitem__(self, index) -> Task:
        ...

    def __add__(self, other: "MetaDataset") -> "MetaConcatDataset":
        return MetaConcatDataset([self, other])


class MetaConcatDataset(MetaDataset):
    datasets: list[MetaDataset]

    def __init__(self, datasets: Iterable[MetaDataset]) -> None:
        super().__init__()
        self.datasets = list(datasets)
        assert len(self.datasets) > 0

    def __getitem__(self, index: tuple[int, Any]) -> Task:
        outer_index, inner_index = index
        return self.datasets[outer_index][inner_index]


class IterableMetaDataset(IterableDataset[Task], ABC):
    task_source: TaskSource

    @abstractmethod
    def __iter__(self) -> Iterator[Task]:
        ...

    def __add__(self, other: "IterableMetaDataset") -> "MetaChainDataset":
        return MetaChainDataset([self, other])


class MetaChainDataset(ChainDataset, IterableMetaDataset, ABC):
    pass



from .cross_domain import CrossDomainModel, SimpleCrossDomainClassifier
from .episodic import EpisodicModel
from .maml import MAML, CrossDomainMAML
from .protonet import ProtoNet, SimpleCrossDomainProtoNet
from .simple import SimpleClassifier



from typing import Callable

import torch
import torchopt
from torch import nn


def inner_loop(
    net: nn.Module,
    optim: torchopt.Optimizer,
    num_steps: int,
    support_x: torch.Tensor,
    support_y: torch.Tensor,
    loss_func: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
) -> None:
    for _ in range(num_steps):
        support_logit = net(support_x)
        loss = loss_func(support_logit, support_y)
        optim.zero_grad()  # zero gradients
        loss.backward()  # backward
        optim.step()  # step updates
    optim.zero_grad()  # zero gradients


def differentiable_inner_loop(
    net: nn.Module,
    optim: torchopt.MetaOptimizer,
    num_steps: int,
    support_x: torch.Tensor,
    support_y: torch.Tensor,
    loss_func: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
) -> None:
    for _ in range(num_steps):
        support_logit = net(support_x)
        loss = loss_func(support_logit, support_y)
        optim.step(loss)



import random
from collections.abc import Sequence, Iterator
from functools import partial
from typing import TypeVar

import numpy as np
import torch

from torchcross.utils.collate_fn import identity
from .base import IterableMetaDataset
from ..task import TaskTarget, Task, TaskDescription
from ..task_source import TaskSource

LT = TypeVar("LT", bound=torch.Tensor | np.ndarray | Sequence[int])


def get_indices(
    labels: LT, classes, task_target: TaskTarget
) -> tuple[dict[int, LT], dict[int, LT]]:
    if isinstance(labels, np.ndarray):

        def ind_func(x):
            return np.nonzero(x)[0]

    elif isinstance(labels, torch.Tensor):

        def ind_func(x):
            return torch.nonzero(x).flatten()

    else:
        raise ValueError("Unsupported label data type")

    if task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
        pos_indices = {c: ind_func(labels == c) for c in classes}
        neg_indices = {c: ind_func(labels != c) for c in classes}
    elif task_target is TaskTarget.MULTILABEL_CLASSIFICATION:
        pos_indices = {c: ind_func(labels[:, c]) for c in classes}
        neg_indices = {c: ind_func(labels[:, c] - 1) for c in classes}
    elif task_target is TaskTarget.BINARY_CLASSIFICATION:
        pos_indices = {1: ind_func(labels)}
        neg_indices = {1: ind_func(labels - 1)}
    elif task_target is TaskTarget.ORDINAL_REGRESSION:
        pos_indices = {c: ind_func(labels == c) for c in classes}
        neg_indices = {c: ind_func(labels != c) for c in classes}
    else:
        raise NotImplementedError(f"Task target {task_target} not yet implemented")
    return pos_indices, neg_indices


def multiclass_few_shot_sample(
    pos_indices: dict[int, LT],
    n_shot: int,
    selected_classes: Sequence[int] | dict[int, str],
) -> np.ndarray:
    sample = np.array([], dtype=np.int64)
    # Sample n_shot positive examples for each class
    for c in selected_classes:
        pos_sample = np.random.choice(pos_indices[c], size=n_shot, replace=False)
        sample = np.concatenate([sample, pos_sample])
    return sample


def binary_few_shot_sample(
    pos_indices: dict[int, LT], neg_indices: dict[int, LT], n_shot: int
) -> np.ndarray:
    pos_sample = np.random.choice(pos_indices[1], size=n_shot, replace=False)
    neg_sample = np.random.choice(neg_indices[1], size=n_shot, replace=False)
    return np.concatenate([pos_sample, neg_sample])


def multilabel_few_shot_sample(
    labels: np.ndarray | torch.Tensor,
    pos_indices: dict[int, LT],
    neg_indices: dict[int, LT],
    n_shot: int,
    selected_classes: Sequence[int] | dict[int, str],
) -> np.ndarray:
    if isinstance(labels, np.ndarray):
        sum_func = partial(np.sum, axis=0)
        all_func = np.all
    elif isinstance(labels, torch.Tensor):
        sum_func = partial(torch.sum, dim=0)
        all_func = torch.all
    else:
        raise ValueError("Unsupported data type")

    if isinstance(selected_classes, dict):
        selected_classes = list(selected_classes.keys())
    labels = labels[:, selected_classes]

    def criterion(sample_to_test: np.ndarray) -> bool:
        column_sums = sum_func(labels[sample_to_test])
        pos_criterion = all_func(column_sums >= n_shot)
        neg_criterion = all_func(column_sums <= len(sample_to_test) - n_shot)
        return pos_criterion and neg_criterion

    sample = np.array([], dtype=np.int64)

    # Sample n_shot positive and n_shot negative examples for each label
    for c in selected_classes:
        pos_sample = np.random.choice(pos_indices[c], size=n_shot, replace=False)
        neg_sample = np.random.choice(neg_indices[c], size=n_shot, replace=False)
        sample = np.concatenate([sample, pos_sample, neg_sample])

    # Remove duplicates and shuffle the sample for the later steps
    sample = np.unique(sample)
    np.random.shuffle(sample)

    # Add single examples from the sample until criterion is satisfied, i.e. until
    # there are enough positive and negative examples for each label.
    # We do this so we have to remove fewer examples (and therefore need to compute the
    # criterion fewer times) later on.
    for i in range(2 * n_shot, len(sample)):
        small_sample = sample[:i]
        if criterion(small_sample):
            sample = small_sample
            break

    # Remove examples (greedily) as long as criterion is satisfied without the example,
    # i.e. as long as there are enough positive and negative examples left for each
    # label.
    i = 0
    while i < len(sample):
        row_removed = np.delete(sample, i)
        if criterion(row_removed):
            sample = row_removed
        else:
            i += 1

    return sample


def remove_indices(
    pos_indices: dict[int, LT],
    neg_indices: dict[int, LT],
    sample: LT,
) -> tuple[dict[int, LT], dict[int, LT]]:
    sample_set = set(sample)

    def _compute(class_indices):
        if isinstance(class_indices, np.ndarray):
            return np.setdiff1d(class_indices, sample)
        elif isinstance(class_indices, torch.Tensor):
            return torch.from_numpy(np.setdiff1d(class_indices, sample))
            # return class_indices[
            #     torch.isin(class_indices, sample, assume_unique=True, invert=True)
            # ]
        else:
            return [i for i in class_indices if i not in sample_set]

    new_pos_indices = {c: _compute(c_indices) for c, c_indices in pos_indices.items()}
    new_neg_indices = {c: _compute(c_indices) for c, c_indices in neg_indices.items()}
    return new_pos_indices, new_neg_indices


class FewShotMetaDataset(IterableMetaDataset):
    def __init__(
        self,
        task_source: TaskSource,
        collate_fn=None,
        n_support_samples_per_class: int = 5,
        n_query_samples_per_class: int = 20,
        filter_classes_min_samples: int = 50,
        length: int | None = None,
        output_indices: bool = False,
    ) -> None:
        super().__init__()
        self.task_source = task_source
        self.task_description = task_source.task_description
        self.task_target = self.task_description.task_target

        self.pos_indices, self.neg_indices = get_indices(
            self.task_source.labels, self.task_description.classes, self.task_target
        )

        self.used_classes = list(self.task_description.classes.keys())
        if self.task_target is TaskTarget.BINARY_CLASSIFICATION:
            self.used_classes = [1]
        if filter_classes_min_samples > 0:
            self.used_classes = [
                c
                for c in self.used_classes
                if len(self.pos_indices[c]) >= filter_classes_min_samples
            ]

            if self.task_target in (
                TaskTarget.BINARY_CLASSIFICATION,
                TaskTarget.MULTILABEL_CLASSIFICATION,
            ):
                self.used_classes = [
                    c
                    for c in self.used_classes
                    if len(self.neg_indices[c]) >= filter_classes_min_samples
                ]
        if (
            len(self.used_classes) < 2
            and self.task_target is not TaskTarget.BINARY_CLASSIFICATION
        ):
            message = "FewShotMetaDataset requires at least 2 classes"
            if filter_classes_min_samples > 0:
                message += f" with at least {filter_classes_min_samples} samples each"
            message += f"for {self.task_target}."
            raise ValueError(message)

        self.collate_fn = collate_fn if collate_fn else identity
        self.n_support_samples_per_class = n_support_samples_per_class
        self.n_query_samples_per_class = n_query_samples_per_class
        self.length = length
        self.output_indices = output_indices

    def __len__(self) -> int:
        if self.length is None:
            raise ValueError("Length is not set.")
        return self.length

    def __iter__(self) -> Iterator[Task]:
        match self.task_target:
            case TaskTarget.MULTICLASS_CLASSIFICATION | TaskTarget.ORDINAL_REGRESSION:
                return self._multiclass_iter()
            case TaskTarget.MULTILABEL_CLASSIFICATION:
                return self._multilabel_iter()
            case TaskTarget.BINARY_CLASSIFICATION:
                return self._binary_iter()
            case target:
                msg = f"Task target {target} is not supported by FewShotMetaDataset."
                raise NotImplementedError(msg)

    def _multiclass_iter(self) -> Iterator[Task]:
        pos_indices = self.pos_indices
        neg_indices = self.neg_indices
        iter_index = 0
        stop = False
        while not stop:
            try:
                support_indices = multiclass_few_shot_sample(
                    pos_indices, self.n_support_samples_per_class, self.used_classes
                )
                new_pos_indices, new_neg_indices = remove_indices(
                    pos_indices, neg_indices, support_indices
                )
                query_indices = multiclass_few_shot_sample(
                    new_pos_indices, self.n_query_samples_per_class, self.used_classes
                )
                if self.length is None:
                    pos_indices, neg_indices = remove_indices(
                        new_pos_indices, new_neg_indices, query_indices
                    )
                else:
                    iter_index += 1
                    if iter_index >= self.length:
                        stop = True
                class_to_label = {
                    c: torch.tensor(ci) for ci, c in enumerate(self.used_classes)
                }
                support = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        class_to_label[self.task_source.labels[i].item()],
                    )
                    for i in support_indices
                ]
                query = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        class_to_label[self.task_source.labels[i].item()],
                    )
                    for i in query_indices
                ]
                support = self.collate_fn(support)
                query = self.collate_fn(query)
                class_dict = {
                    ci: self.task_description.classes[c]
                    for ci, c in enumerate(self.used_classes)
                }
                yield Task(
                    support, query, TaskDescription(self.task_target, class_dict)
                )
            except ValueError:
                return

    def _multilabel_iter(self) -> Iterator[Task]:
        pos_indices = self.pos_indices
        neg_indices = self.neg_indices
        iter_index = 0
        stop = False
        while not stop:
            try:
                support_indices = multilabel_few_shot_sample(
                    self.task_source.labels,
                    pos_indices,
                    neg_indices,
                    self.n_support_samples_per_class,
                    self.used_classes,
                )
                new_pos_indices, new_neg_indices = remove_indices(
                    pos_indices, neg_indices, support_indices
                )
                query_indices = multilabel_few_shot_sample(
                    self.task_source.labels,
                    new_pos_indices,
                    new_neg_indices,
                    self.n_query_samples_per_class,
                    self.used_classes,
                )
                if self.length is None:
                    pos_indices, neg_indices = remove_indices(
                        new_pos_indices, new_neg_indices, query_indices
                    )
                else:
                    iter_index += 1
                    if iter_index >= self.length:
                        stop = True

                labels = self.task_source.labels[:, self.used_classes]
                support = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        torch.tensor(labels[i]),
                    )
                    for i in support_indices
                ]
                query = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        torch.tensor(labels[i]),
                    )
                    for i in query_indices
                ]
                support = self.collate_fn(support)
                query = self.collate_fn(query)
                class_dict = {
                    ci: self.task_description.classes[c]
                    for ci, c in enumerate(self.used_classes)
                }
                yield Task(
                    support, query, TaskDescription(self.task_target, class_dict)
                )
            except ValueError:
                return

    def _binary_iter(self) -> Iterator[Task]:
        pos_indices = self.pos_indices
        neg_indices = self.neg_indices
        iter_index = 0
        stop = False
        while not stop:
            try:
                support_indices = binary_few_shot_sample(
                    pos_indices, neg_indices, self.n_support_samples_per_class
                )
                new_pos_indices, new_neg_indices = remove_indices(
                    pos_indices, neg_indices, support_indices
                )
                query_indices = binary_few_shot_sample(
                    new_pos_indices, new_neg_indices, self.n_query_samples_per_class
                )
                if self.length is None:
                    pos_indices, neg_indices = remove_indices(
                        new_pos_indices, new_neg_indices, query_indices
                    )
                else:
                    iter_index += 1
                    if iter_index >= self.length:
                        stop = True

                labels = self.task_source.labels
                support = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        torch.tensor(labels[i]),
                    )
                    for i in support_indices
                ]
                query = [
                    (
                        i if self.output_indices else self.task_source[i][0],
                        torch.tensor(labels[i]),
                    )
                    for i in query_indices
                ]
                support = self.collate_fn(support)
                query = self.collate_fn(query)
                class_dict = {
                    0: self.task_description.classes[0],
                    1: self.task_description.classes[1],
                }
                yield Task(
                    support, query, TaskDescription(self.task_target, class_dict)
                )

            except ValueError:
                return


class SubTaskRandomFewShotMetaDataset(FewShotMetaDataset):
    def __init__(
        self,
        task_source: TaskSource,
        collate_fn=None,
        *,
        n_support_samples_per_class_min: int = 1,
        n_support_samples_per_class_max: int = 10,
        n_query_samples_per_class: int = 10,
        filter_classes_min_samples: int = 40,
        length: int | None = None,
        output_indices: bool = False,
    ) -> None:
        super().__init__(
            task_source,
            collate_fn,
            n_support_samples_per_class_min,
            n_query_samples_per_class,
            filter_classes_min_samples,
            length,
            output_indices,
        )
        self.original_used_classes = self.used_classes
        self.n_support_samples_per_class_min = n_support_samples_per_class_min
        self.n_support_samples_per_class_max = n_support_samples_per_class_max

    def __iter__(self):
        if self.task_target != TaskTarget.BINARY_CLASSIFICATION:
            n_way = random.randint(2, len(self.original_used_classes))
            self.used_classes = random.sample(self.original_used_classes, n_way)
        self.n_support_samples_per_class = random.randint(
            self.n_support_samples_per_class_min, self.n_support_samples_per_class_max
        )
        for task in super().__iter__():
            yield task
            if self.task_target != TaskTarget.BINARY_CLASSIFICATION:
                n_way = random.randint(2, len(self.original_used_classes))
                self.used_classes = random.sample(self.original_used_classes, n_way)
            self.n_support_samples_per_class = random.randint(
                self.n_support_samples_per_class_min,
                self.n_support_samples_per_class_max,
            )
        self.used_classes = self.original_used_classes



import torch

from torchcross.utils.collate_fn import identity
from . import MetaDataset
from ..task_source import TaskSource
from ..task import TaskTarget, Task, TaskDescription


class TakeFirstFewShotMetaDataset(MetaDataset):
    def __init__(
        self,
        dataset: TaskSource,
        collate_fn=None,
        n_shot_per_class: int = 5,
    ) -> None:
        super().__init__()
        self.dataset = dataset
        self.task_description = dataset.task_description
        self.task_target = self.task_description.task_target

        if self.task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
            self.indices = {
                c: [i for i, l in enumerate(self.dataset.labels) if l == c]
                for c in self.task_description.classes
            }
        else:
            raise NotImplementedError(
                f"Task target {self.task_target} not yet implemented"
            )

        def multiclass_n_shot(start=0):
            sample = [
                (i, j)
                for j, c in enumerate(self.task_description.classes)
                for i in self.indices[c][start : start + n_shot_per_class]
            ]
            return [(self.dataset[i][0], torch.tensor(j)) for i, j in sample]

        self.collate_fn = collate_fn if collate_fn else identity

        self.support = self.collate_fn(multiclass_n_shot(0))
        self.query = self.collate_fn(multiclass_n_shot(n_shot_per_class))

    def __getitem__(self, _) -> Task:
        return Task(
            self.support,
            self.query,
            TaskDescription(self.task_target, self.task_description.classes),
        )



from typing import Sequence

import torch
from lightning import pytorch as pl
from torch import Tensor

from torchcross import models
from torchcross.data.task import Task
from torchcross.models.lightning import mixins


class EpisodicLightningModule(
    models.EpisodicModel, mixins.MeanMetricsMixin, pl.LightningModule
):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.automatic_optimization = False

    def get_metrics_and_losses(
        self, batch: Sequence[Task]
    ) -> tuple[list[tuple[Tensor, ...]], list[Tensor], list[Tensor]]:
        return [], [], self.forward(batch)

    def training_step(self, batch: Sequence[Task], batch_idx: int):
        self.optimizers().zero_grad()

        metric_values, support_losses, query_losses = self.get_metrics_and_losses(batch)

        self.update_metrics("train", metric_values)
        support_loss = torch.stack(support_losses).mean()
        query_loss = torch.stack(query_losses).mean()

        self.manual_backward(query_loss)
        self.optimizers().step()

        self.log("loss/metatrain/support", support_loss, batch_size=len(batch))
        self.log(
            "loss/metatrain/query", query_loss, batch_size=len(batch), prog_bar=True
        )
        self.log_dict(self.training_metrics)

    def validation_step(self, batch: Sequence[Task], batch_idx: int):
        torch.set_grad_enabled(True)
        metric_values, support_losses, query_losses = self.get_metrics_and_losses(batch)
        torch.set_grad_enabled(False)

        self.update_metrics("val", metric_values)
        support_loss = torch.stack(support_losses).mean()
        query_loss = torch.stack(query_losses).mean()

        self.log("loss/metaval/support", support_loss, batch_size=len(batch))
        self.log("loss/metaval/query", query_loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.validation_metrics)

    def test_step(self, batch: Sequence[Task], batch_idx: int):
        torch.set_grad_enabled(True)
        metric_values, support_losses, query_losses = self.get_metrics_and_losses(batch)
        torch.set_grad_enabled(False)

        self.update_metrics("test", metric_values)
        support_loss = torch.stack(support_losses).mean()
        query_loss = torch.stack(query_losses).mean()

        self.log("loss/metatest/support", support_loss, batch_size=len(batch))
        self.log(
            "loss/metatest/query", query_loss, batch_size=len(batch), prog_bar=True
        )
        self.log_dict(self.test_metrics)

    def configure_optimizers(self):
        return self.outer_optimizer(self.parameters())



import pytest
from typing import List
from torch.utils.data import Dataset, TensorDataset
import torch
from torchcross.data import TaskDescription, Task, TaskTarget
from torchcross.data.metadataset import (
    CollectionMetaDataset,
)  # replace with actual import


def create_task() -> Task:
    return Task(
        create_dataset(),
        create_dataset(),
        TaskDescription(TaskTarget.MULTICLASS_CLASSIFICATION, {0: "a", 1: "b", 2: "c"}),
    )


def create_dataset() -> Dataset:
    return TensorDataset(torch.rand(10), torch.randint(0, 3, (10,)))


def create_task_description() -> TaskDescription:
    return TaskDescription(
        TaskTarget.MULTICLASS_CLASSIFICATION,
        {0: "a", 1: "b"},
        "test",
    )


class TestCollectionMetaDataset:
    @pytest.fixture
    def tasks(self) -> List[Task]:
        return [create_task() for _ in range(5)]

    @pytest.fixture
    def datasets(self) -> List[Dataset]:
        return [create_dataset() for _ in range(5)]

    @pytest.fixture
    def task_descriptions(self) -> List[TaskDescription]:
        return [create_task_description() for _ in range(5)]

    def test_init_with_tasks(self, tasks):
        dataset = CollectionMetaDataset(tasks)
        assert dataset.tasks == tasks

    def test_init_with_datasets_and_descriptions(self, datasets, task_descriptions):
        dataset = CollectionMetaDataset(datasets, datasets, task_descriptions)
        assert len(dataset.tasks) == len(datasets)
        for task, dataset, task_description in zip(
            dataset.tasks, datasets, task_descriptions
        ):
            assert task.support == dataset
            assert task.query == dataset
            assert task.task_target == task_description.task_target
            assert task.classes == task_description.classes

    def test_init_with_invalid_arguments(self, tasks, datasets):
        with pytest.raises(TypeError):
            CollectionMetaDataset(tasks, datasets, datasets)

    def test_init_with_no_arguments(self):
        with pytest.raises(TypeError):
            CollectionMetaDataset()

    def test_tasks_property(self, tasks):
        dataset = CollectionMetaDataset(tasks)
        assert dataset.tasks == tasks



import torch
import torch.nn as nn


class Expand(nn.Module):
    """Uses `torch.Tensor.expand` to expand a tensor to a given size.

    Args:
        sizes: The sizes to expand the tensor to.
    """

    def __init__(self, *sizes):
        super().__init__()
        self.sizes = sizes

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.expand(*self.sizes)



from .task import Task, TaskDescription, TaskTarget
from .task_source import (
    TaskSource,
    WrapTaskSource,
    BatchedTaskSource,
    ConcatTaskSource,
)
from .dataset import InterleaveDataset, RandomInterleaveDataset

__all__ = [
    "Task",
    "TaskDescription",
    "TaskTarget",
    "TaskSource",
    "WrapTaskSource",
    "BatchedTaskSource",
    "ConcatTaskSource",
    "InterleaveDataset",
    "RandomInterleaveDataset",
]



from typing import Literal

import torch

from torchcross.data.task import TaskTarget


def new_classification_head(
    task_target: TaskTarget,
    classes: dict[int, str],
    num_in_features: int,
    device: torch.device = None,
    init: Literal["default", "kaiming", "zero"] = "zero",
) -> torch.nn.Linear:
    """Create a new classification head for a given task target.

    Args:
        task_target: The task target.
        classes: The classes for the task.
        num_in_features: The number of input features for the Linear layer.
        device: The device to create the Linear layer on.
        init: The initialization method for the Linear layer.

    Returns:
        A new Linear layer for the given task target and classes.
    """
    match task_target:
        case (
            TaskTarget.MULTICLASS_CLASSIFICATION
            | TaskTarget.MULTILABEL_CLASSIFICATION
            | TaskTarget.ORDINAL_REGRESSION
        ):
            fc = torch.nn.Linear(num_in_features, len(classes), device=device)
        case TaskTarget.BINARY_CLASSIFICATION | TaskTarget.REGRESSION:
            fc = torch.nn.Linear(num_in_features, 1, device=device)
        case target:
            raise NotImplementedError(f"Task target {target} is not supported")
    if init == "zero":
        torch.nn.init.zeros_(fc.weight.data)
        torch.nn.init.zeros_(fc.bias.data)
    elif init == "kaiming" or init == "default":
        pass
    else:
        raise ValueError(f"Unknown init method {init}")
    return fc



from collections import OrderedDict

import torch
from torch import nn

from torchcross.cd.heads import new_classification_head
from torchcross.data.task import TaskDescription
from torchcross.utils.layers import Expand


class SimpleClassifier(nn.Module):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        task_description: TaskDescription,
        expand_input_channels: bool = True,
    ) -> None:
        super(SimpleClassifier, self).__init__()

        self.backbone = backbone
        self.num_backbone_features = num_backbone_features
        self.head = new_classification_head(
            task_description.task_target,
            task_description.classes,
            num_backbone_features,
            init="kaiming",
        )
        net_dict = OrderedDict(backbone=self.backbone, head=self.head)
        if expand_input_channels:
            net_dict = OrderedDict(input_module=Expand(-1, 3, -1, -1), **net_dict)
        self.net = torch.nn.Sequential(net_dict)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.net(x)



from pprint import pprint
from typing import Literal

import torch
from lightning import pytorch as pl
from torch import Tensor

from torchcross import models
from torchcross.cd.activations import get_prob_func
from torchcross.cd.losses import get_loss_func
from torchcross.data.task import TaskDescription
from torchcross.models.lightning import mixins


class CrossDomainLightningModule(models.CrossDomainModel, pl.LightningModule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.scale_loss_by_batch_count = False
        self.train_counts = {}
        self.val_counts = {}
        self.test_counts = {}
        self.previous_train_counts = {}
        self.previous_val_counts = {}
        self.previous_test_counts = {}

        self.automatic_optimization = False
        self.debug = False

    def get_metrics_and_loss(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        mode: Literal["train", "val", "test"],
    ) -> tuple[tuple[Tensor, ...], Tensor]:
        (x, y), task_description = batch
        logits = self.forward(x, task_description)
        loss_func = get_loss_func(
            task_description.task_target, task_description.classes, self.device
        )
        pred_func = get_prob_func(task_description.task_target)
        loss = loss_func(logits, y)

        pred = pred_func(logits)
        task_counter = getattr(self, f"{mode}_counts")
        task_counter[task_description.task_identifier] = (
            task_counter.get(task_description.task_identifier, 0) + 1
        )
        return self.call_metrics(mode, task_description, pred, y), loss

    def on_train_epoch_start(self) -> None:
        self.previous_train_counts = self.train_counts
        self.train_counts = {}

    def on_validation_epoch_start(self) -> None:
        self.previous_val_counts = self.val_counts
        self.val_counts = {}

    def on_test_epoch_start(self) -> None:
        self.previous_test_counts = self.test_counts
        self.test_counts = {}

    def on_train_epoch_end(self) -> None:
        if self.debug:
            print("Train counts:")
            pprint(self.train_counts)
            print(len(self.train_counts))
            print(sum(self.train_counts.values()))

    def on_validation_epoch_end(self) -> None:
        if self.debug:
            print("Val counts:")
            pprint(self.val_counts)
            print(len(self.val_counts))
            print(sum(self.val_counts.values()))

    def on_test_epoch_end(self) -> None:
        if self.debug:
            print("Test counts:")
            pprint(self.test_counts)
            print(len(self.test_counts))
            print(sum(self.test_counts.values()))

    def call_metrics(
        self,
        mode: Literal["train", "val", "test"],
        task_description: TaskDescription,
        pred: torch.Tensor,
        y: torch.Tensor,
    ):
        raise NotImplementedError

    def training_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        self.optimizers().zero_grad()

        metric_values, loss = self.get_metrics_and_loss(batch, "train")

        scaled_loss = loss
        if self.scale_loss_by_batch_count:
            # Scale the loss by the count of batches for this task in the previous
            # epoch compared to the task with the most batches in the previous epoch.
            # We use the previous epoch because the current epoch is not finished yet
            # and therefore the counts are not final.
            task_identifier = batch[1].task_identifier
            previous_task_count = self.previous_train_counts.get(task_identifier, 0)
            if previous_task_count:
                # If the task was not present in the previous epoch, we do not scale the
                # loss, as it is the first epoch or the first time we see this task.
                previous_max_count = max(self.previous_train_counts.values())
                scale_factor = previous_max_count / previous_task_count
                scaled_loss = loss * scale_factor

        self.manual_backward(scaled_loss)
        self.optimizers().step()

        self.log("loss/train", loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.training_metrics)

    def validation_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        metric_values, loss = self.get_metrics_and_loss(batch, "val")

        self.log("loss/val", loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.validation_metrics, prog_bar=True)

    def test_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        metric_values, loss = self.get_metrics_and_loss(batch, "test")

        self.log("loss/test", loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.test_metrics)

    def configure_optimizers(self):
        return self.optimizer(self.parameters())



import math
from collections import OrderedDict
from collections.abc import Sequence, Callable, Iterable
from typing import Optional, Literal

import torch
import torchopt
from torch import nn, Tensor

from torchcross.cd.heads import new_classification_head
from torchcross.cd.losses import get_loss_func
from torchcross.data.task import TaskDescription, Task
from torchcross.models.cross_domain import CrossDomainModel
from torchcross.models.episodic import EpisodicModel
from torchcross.utils.layers import Expand
from torchcross.utils.loops import inner_loop, differentiable_inner_loop

head_init_t = Literal[
    "default",
    "kaiming",
    "zero",
    "adaptive",
    "adaptive normal",
    "adaptive uniform",
    "unicorn",
    "proto",
]


class MAML(EpisodicModel):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        task_description: TaskDescription,
        inner_optimizer: Callable[[torch.nn.Module], torchopt.MetaOptimizer],
        eval_inner_optimizer: Callable[
            [Iterable[torch.nn.Parameter]], torchopt.Optimizer
        ],
        num_inner_steps: int,
        eval_num_inner_steps: Optional[int] = None,
        transductive: bool = True,
        reset_head: Optional[head_init_t] = None,
    ) -> None:
        super(MAML, self).__init__()

        self.backbone = backbone
        self.num_backbone_features = num_backbone_features
        self.head = new_classification_head(
            task_description.task_target,
            task_description.classes,
            num_backbone_features,
            init="kaiming",
        )
        self.net = torch.nn.Sequential(
            OrderedDict(backbone=self.backbone, head=self.head)
        )
        self.inner_loss_func = get_loss_func(
            task_description.task_target, task_description.classes
        )

        self.inner_optimizer: torchopt.MetaOptimizer = inner_optimizer(self.net)
        self.eval_inner_optimizer: torchopt.Optimizer = eval_inner_optimizer(
            self.net.parameters()
        )
        self.num_inner_steps = num_inner_steps
        self.eval_num_inner_steps = eval_num_inner_steps or num_inner_steps

        self.transductive = transductive
        self.reset_head_strategy = reset_head

    def forward(self, batch: Sequence[Task]) -> list[torch.Tensor]:
        net_state_dict = torchopt.extract_state_dict(
            self.net, by="reference", detach_buffers=True
        )

        query_logits = []

        for task in batch:
            ql = self.episode(task)
            query_logits.append(ql)
            torchopt.recover_state_dict(self.net, net_state_dict)

        return query_logits

    def episode(
        self, task: Task, with_support_output=False
    ) -> Tensor | tuple[Tensor, Tensor]:
        support_x, support_y = task.support
        query_x = task.query[0]

        if self.reset_head_strategy is not None:
            self.reset_head()
        net = self.net
        net.train()
        if self.training:
            inner_optim_state_dict = torchopt.extract_state_dict(
                self.inner_optimizer, by="reference"
            )
            differentiable_inner_loop(
                net,
                self.inner_optimizer,
                self.num_inner_steps,
                support_x,
                support_y,
                self.inner_loss_func,
            )
            torchopt.recover_state_dict(self.inner_optimizer, inner_optim_state_dict)
        else:
            inner_loop(
                net,
                self.eval_inner_optimizer,
                self.eval_num_inner_steps,
                support_x,
                support_y,
                self.inner_loss_func,
            )
        if not self.transductive:
            net.eval()
        query_logits = net(query_x)
        if with_support_output:
            with torch.no_grad():
                support_logits = net(support_x)
            return support_logits, query_logits
        return query_logits

    def reset_head(self):
        if (
            self.reset_head_strategy == "default"
            or self.reset_head_strategy == "kaiming"
        ):
            self.head.reset_parameters()
        elif self.reset_head_strategy == "zero":
            torch.nn.init.zeros_(self.head.weight)
            torch.nn.init.zeros_(self.head.bias)
        elif (
            self.reset_head_strategy == "adaptive"
            or self.reset_head_strategy == "adaptive normal"
        ):
            torch.nn.init.normal_(
                self.head.weight,
                self.head.weight.mean().item(),
                self.head.weight.std().item(),
            )
            torch.nn.init.normal_(
                self.head.bias,
                self.head.bias.mean().item(),
                self.head.bias.std().item(),
            )
        elif self.reset_head_strategy == "adaptive uniform":
            weigth_mean = self.head.weight.mean().item()
            weight_std = self.head.weight.std().item()
            wa = weigth_mean - math.sqrt(3) * weight_std
            wb = weigth_mean + math.sqrt(3) * weight_std
            torch.nn.init.uniform_(self.head.weight, wa, wb)
            bias_mean = self.head.bias.mean().item()
            bias_std = self.head.bias.std().item()
            ba = bias_mean - math.sqrt(3) * bias_std
            bb = bias_mean + math.sqrt(3) * bias_std
            torch.nn.init.uniform_(self.head.bias, ba, bb)
        elif self.reset_head_strategy == "unicorn":
            raise NotImplementedError
        elif self.reset_head_strategy == "proto":
            raise NotImplementedError
        else:
            raise ValueError(f"Head reset option {self.reset_head_strategy} unknown.")


class CrossDomainMAML(EpisodicModel, CrossDomainModel):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        inner_optimizer,
        eval_inner_optimizer,
        num_inner_steps: int,
        eval_num_inner_steps: Optional[int] = None,
        transductive: bool = True,
        head_init: Optional[head_init_t] = None,
    ) -> None:
        super(CrossDomainMAML, self).__init__()

        self.backbone = backbone
        self.num_backbone_features = num_backbone_features

        self.inner_optim_constructor = inner_optimizer
        self.eval_inner_optim_constructor = eval_inner_optimizer
        self.num_inner_steps = num_inner_steps
        self.eval_num_inner_steps = eval_num_inner_steps or num_inner_steps

        self.transductive = transductive
        self.head_init = head_init or "zero"
        self.input_module = Expand(-1, 3, -1, -1)

    def get_net(self, task_description: TaskDescription) -> torch.nn.Module:
        return torch.nn.Sequential(
            OrderedDict(
                input_module=self.input_module,
                backbone=self.backbone,
                head=new_classification_head(
                    task_description.task_target,
                    task_description.classes,
                    self.num_backbone_features,
                    device=self.device_dummy.device,
                    init=self.head_init,
                ),
            )
        )

    def forward(self, batch: Sequence[Task]) -> list[torch.Tensor]:
        backbone_state_dict = torchopt.extract_state_dict(
            self.backbone, by="reference", detach_buffers=True
        )

        query_logits = []

        for task in batch:
            query_logit = self.episode(task)
            query_logits.append(query_logit)
            torchopt.recover_state_dict(self.backbone, backbone_state_dict)

        return query_logits

    def episode(
        self, task: Task, with_support_output=False
    ) -> Tensor | tuple[Tensor, Tensor]:
        support_x, support_y = task.support
        query_x = task.query[0]

        net = self.get_net(task.description)
        inner_loss_func = get_loss_func(
            task.description.task_target,
            task.description.classes,
            self.device_dummy.device,
        )
        net.train()
        if self.training:
            differentiable_inner_loop(
                net,
                self.inner_optim_constructor(net),
                self.num_inner_steps,
                support_x,
                support_y,
                inner_loss_func,
            )
        else:
            inner_loop(
                net,
                self.eval_inner_optim_constructor(net.parameters()),
                self.eval_num_inner_steps,
                support_x,
                support_y,
                inner_loss_func,
            )
        if not self.transductive:
            net.eval()
        query_logits = net(query_x)
        if with_support_output:
            with torch.no_grad():
                support_logits = net(support_x)
            return support_logits, query_logits
        return query_logits



from typing import Any

from torch import nn
from torchmetrics import Metric
from torchmetrics.wrappers.abstract import WrapperMetric

from torchcross.data import TaskDescription


class CrossDomainWrapper(WrapperMetric):
    """Wraps around a metric constructor and applies it to each task in a cross-domain setting.

    Args:
        metric_constructor: The class of the metric to wrap around.
        task_descriptions: The task descriptions for each task.
        metric_args: The arguments to pass to the metric constructor.
        metric_kwargs: The keyword arguments to pass to the metric constructor.
    """

    is_differentiable = False

    def __init__(
        self,
        metric_constructor: type[Metric],
        task_descriptions: list[TaskDescription] | None = None,
        metric_args: tuple[Any] = (),
        metric_kwargs: dict[str, Any] | None = None,
        auto_add_metrics: bool = True,
    ) -> None:
        super().__init__()
        self.metric_constructor = metric_constructor
        self.task_descriptions = task_descriptions
        self.metric_args = metric_args
        self.metric_kwargs = metric_kwargs or {}
        self.auto_add_metrics = auto_add_metrics

        self.metrics = nn.ModuleDict()
        if task_descriptions is not None:
            for task_description in task_descriptions:
                self.add_metric(task_description)

    def add_metric(self, task_description: TaskDescription) -> None:
        """Add a metric for a new task.

        Args:
            task_description: The task description for the new task.
        """
        task_identifier = task_description.task_identifier
        if task_identifier in self.metrics:
            raise ValueError(f"Task {task_identifier} already has a metric.")
        self.metrics[task_identifier] = self.metric_constructor(
            task_description,
            *self.metric_args,
            **self.metric_kwargs,
        ).to(self.device)

    def update(self, task_description: TaskDescription, *args, **kwargs) -> None:
        """Update the metric for a task.

        Args:
            task_description: The task description for the task.
            args: The positional arguments to pass to the metric.
            kwargs: The keyword arguments to pass to the metric.
        """
        task_identifier = task_description.task_identifier
        if task_identifier not in self.metrics:
            if self.auto_add_metrics:
                self.add_metric(task_description)
            else:
                raise ValueError(f"Task {task_identifier} does not have a metric.")
        self.metrics[task_identifier].update(*args, **kwargs)

    def compute(self) -> Any:
        """Compute the metric for each task.

        Returns:
            The computed metric for each task.
        """
        return {
            task_identifier: metric.compute()
            for task_identifier, metric in self.metrics.items()
        }

    def forward(self, task_description: TaskDescription, *args, **kwargs) -> Any:
        """Forward input to the metric for a task.

        Args:
            task_description: The task description for the task.
            args: The positional arguments to pass to the metric.
            kwargs: The keyword arguments to pass to the metric.

        Returns:
            The output of the metric for the task.
        """
        task_identifier = task_description.task_identifier
        if task_identifier not in self.metrics:
            if self.auto_add_metrics:
                self.add_metric(task_description)
            else:
                raise ValueError(f"Task {task_identifier} does not have a metric.")
        self._forward_cache = self.metrics[task_identifier](*args, **kwargs)
        return self._forward_cache

    def reset(self) -> None:
        """Reset all underlying metrics."""
        for metric in self.metrics.values():
            metric.reset()
        super().reset()


class CrossDomainMeanWrapper(CrossDomainWrapper):
    """Wraps around a metric constructor and applies it to each task in a cross-domain setting.
    Computes the mean of the metric across tasks.
    """

    def compute(self) -> Any:
        """Compute the mean of the metric over all tasks.

        Returns:
            The mean of the computed metric over all tasks.
        """
        return sum(m.compute() for m in self.metrics.values()) / len(self.metrics)



from .base import MetaDataset, MetaConcatDataset, IterableMetaDataset, MetaChainDataset
from .few_shot import (
    FewShotMetaDataset,
    SubTaskRandomFewShotMetaDataset,
)
from .few_shot_dummies import TakeFirstFewShotMetaDataset
from .collection import CollectionMetaDataset

__all__ = [
    "MetaDataset",
    "MetaConcatDataset",
    "IterableMetaDataset",
    "MetaChainDataset",
    "FewShotMetaDataset",
    "SubTaskRandomFewShotMetaDataset",
    "TakeFirstFewShotMetaDataset",
    "CollectionMetaDataset",
]



from functools import partial
from typing import Callable

import torch
from torch.nn import functional as F

from torchcross.data.task import TaskTarget

__all__ = ["get_prob_func", "get_log_prob_func"]


def get_prob_func(
    task_target: TaskTarget,
) -> Callable[[torch.Tensor], torch.Tensor]:
    """Get the probability function for a given task target.

    Args:
        task_target: The task target.

    Returns:
        A function to convert logits to probabilities.
    """
    match task_target:
        case TaskTarget.MULTICLASS_CLASSIFICATION:
            return partial(torch.softmax, dim=-1)
        case TaskTarget.MULTILABEL_CLASSIFICATION | TaskTarget.BINARY_CLASSIFICATION:
            return torch.sigmoid
        case TaskTarget.ORDINAL_REGRESSION:
            return partial(torch.softmax, dim=-1)
        case TaskTarget.REGRESSION:
            return lambda x: x
        case target:
            raise NotImplementedError(f"Task target {target} not yet implemented")


def get_log_prob_func(
    task_target: TaskTarget,
) -> Callable[[torch.Tensor], torch.Tensor]:
    """Get the log probability function for a given task target.

    Args:
        task_target: The task target.

    Returns:
        A function to convert logits to log probabilities.
    """
    match task_target:
        case TaskTarget.MULTICLASS_CLASSIFICATION:
            return partial(torch.log_softmax, dim=-1)
        case TaskTarget.MULTILABEL_CLASSIFICATION | TaskTarget.BINARY_CLASSIFICATION:
            return F.logsigmoid
        case TaskTarget.ORDINAL_REGRESSION:
            return partial(torch.log_softmax, dim=-1)
        case TaskTarget.REGRESSION:
            return lambda x: x
        case target:
            raise NotImplementedError(f"Task target {target} not yet implemented")



from typing import Sequence

import torch
from torch import nn

from torchcross.data import Task


class EpisodicModel(nn.Module):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def forward(self, batch: Sequence[Task]) -> list[torch.Tensor]:
        return [self.episode(task) for task in batch]

    def episode(self, task: Task) -> torch.Tensor:
        ...



from typing import Callable, Any, Literal

import torch
from lightning import pytorch as pl
from torch import Tensor, nn
from torch.optim import Optimizer
from torchmetrics import MetricCollection, Metric

from torchcross import models
from torchcross.cd.activations import get_prob_func
from torchcross.cd.losses import get_loss_func, get_criterion
from torchcross.cd.metrics import get_accuracy_func, get_auroc_func, Accuracy, AUROC
from torchcross.data.task import TaskDescription
from torchcross.models.lightning import mixins


class SimpleClassifier(models.SimpleClassifier, pl.LightningModule):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        task_description: TaskDescription,
        optimizer: type[Optimizer],
        lr_scheduler: Callable[[Optimizer], Any] | None = None,
        expand_input_channels: bool = True,
        pos_class_weights: torch.Tensor = None,
    ) -> None:
        super().__init__(
            backbone, num_backbone_features, task_description, expand_input_channels
        )

        self.register_buffer("pos_class_weights", pos_class_weights)
        self.criterion = get_criterion(
            task_description, pos_class_weights=pos_class_weights
        )
        self.pred_func = get_prob_func(task_description.task_target)

        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

        self.training_metrics = MetricCollection(
            {"accuracy": Accuracy(task_description), "AUROC": AUROC(task_description)},
            postfix="/train",
        )
        self.validation_metrics = self.training_metrics.clone(postfix="/val")
        self.test_metrics = self.training_metrics.clone(postfix="/test")

        self.automatic_optimization = False

    def do_step(
        self,
        batch: tuple[torch.Tensor, torch.Tensor],
        mode: Literal["train", "val", "test"],
    ) -> Tensor:
        x, y = batch
        logits = self.forward(x)
        # loss = self.loss_func(logits, y, self.pos_class_weights)
        loss = self.criterion(logits, y)
        pred = self.pred_func(logits)
        self.update_metrics(mode, pred, y)
        return loss

    def update_metrics(
        self, mode: Literal["train", "val", "test"], pred: torch.Tensor, y: torch.Tensor
    ):
        metrics = self.get_metrics(mode)
        return metrics(pred, y)

    def get_metrics(
        self, mode: Literal["train", "val", "test"]
    ) -> Metric | MetricCollection:
        if mode == "train":
            return self.training_metrics
        elif mode == "val":
            return self.validation_metrics
        elif mode == "test":
            return self.test_metrics
        else:
            raise ValueError(f"Invalid mode: {mode}")

    def training_step(
        self,
        batch: tuple[torch.Tensor, torch.Tensor],
        batch_idx: int,
    ):
        self.optimizers().zero_grad()

        loss = self.do_step(batch, "train")

        self.manual_backward(loss)
        self.optimizers().step()

        self.log("loss/train", loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.training_metrics)

    def validation_step(
        self,
        batch: tuple[torch.Tensor, torch.Tensor],
        batch_idx: int,
    ):
        loss = self.do_step(batch, "val")

        self.log("loss/val", loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.validation_metrics, prog_bar=True)

    def test_step(
        self,
        batch: tuple[torch.Tensor, torch.Tensor],
        batch_idx: int,
    ):
        loss = self.do_step(batch, "test")

        self.log("loss/test", loss, batch_size=len(batch), prog_bar=True)
        self.log_dict(self.test_metrics)

    def configure_optimizers(self):
        return self.optimizer(self.parameters())



from collections import OrderedDict

import torch
from torch import nn

from torchcross.cd.heads import new_classification_head
from torchcross.data.task import TaskDescription
from torchcross.utils.layers import Expand


class CrossDomainModel(nn.Module):
    def __init__(self) -> None:
        super(CrossDomainModel, self).__init__()
        self.register_buffer("device_dummy", torch.empty(0))

    def forward(
        self, x: torch.Tensor, task_description: TaskDescription
    ) -> torch.Tensor:
        net_for_task = self.get_net(task_description)
        return net_for_task(x)

    def get_net(self, task_description: TaskDescription) -> torch.nn.Module:
        raise NotImplementedError


class SimpleCrossDomainClassifier(CrossDomainModel):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        task_descriptions: list[TaskDescription] = None,
        add_heads_during_training: bool = True,
    ) -> None:
        super(SimpleCrossDomainClassifier, self).__init__()

        self.backbone = backbone
        self.num_backbone_features = num_backbone_features

        self.heads = torch.nn.ModuleDict()
        if task_descriptions is None:
            task_descriptions = []
        for task_description in task_descriptions:
            self.add_head(task_description)

        self.add_heads_during_training = add_heads_during_training

    def get_net(self, task_description: TaskDescription) -> torch.nn.Module:
        if task_description.task_identifier not in self.heads:
            if self.training and self.add_heads_during_training:
                self.add_head(task_description)
            else:
                raise ValueError(
                    f"Head for task {task_description.task_identifier} does not exist."
                )
        return torch.nn.Sequential(
            OrderedDict(
                input_module=Expand(-1, 3, -1, -1),
                backbone=self.backbone,
                head=self.heads[task_description.task_identifier],
            )
        )

    def add_head(self, task_description: TaskDescription) -> nn.Module:
        task_identifier = task_description.task_identifier
        if task_identifier in self.heads:
            raise ValueError(f"Head for task {task_identifier} already exists.")
        self.heads[task_identifier] = new_classification_head(
            task_description.task_target,
            task_description.classes,
            self.num_backbone_features,
            self.device_dummy.device,
        )
        return self.heads[task_identifier]



from typing import Literal

import torch
import torchmetrics
from torch import Tensor
from lightning import pytorch as pl


class MeanMetricsMixin(pl.LightningModule):
    def __init__(self) -> None:
        super().__init__()
        self.metric_keys = []
        self.training_metrics = torch.nn.ModuleDict()
        self.validation_metrics = torch.nn.ModuleDict()
        self.test_metrics = torch.nn.ModuleDict()

    def configure_metrics(self, metric_keys: list[str]) -> None:
        self.metric_keys = metric_keys
        self.training_metrics = torch.nn.ModuleDict(
            {k.format("train"): torchmetrics.MeanMetric() for k in self.metric_keys}
        )
        self.validation_metrics = torch.nn.ModuleDict(
            {k.format("val"): torchmetrics.MeanMetric() for k in self.metric_keys}
        )
        self.test_metrics = torch.nn.ModuleDict(
            {k.format("test"): torchmetrics.MeanMetric() for k in self.metric_keys}
        )

    def update_metrics(
        self,
        mode: Literal["train", "val", "test"],
        batch_metrics: list[tuple[Tensor, ...]] | tuple[Tensor, ...],
    ):
        metrics: torch.nn.ModuleDict
        if mode == "train":
            metrics = self.training_metrics
        elif mode == "val":
            metrics = self.validation_metrics
        elif mode == "test":
            metrics = self.test_metrics
        else:
            raise ValueError(f"Unknown mode {mode}")

        match batch_metrics:
            case tuple():
                aggregated = batch_metrics
                batch_size = 1
            case list():
                aggregated = [torch.stack(bm).mean() for bm in zip(*batch_metrics)]
                batch_size = len(batch_metrics)
            case _:
                raise ValueError(f"Unknown batch_metrics type {type(batch_metrics)}")

        for k, v in zip(self.metric_keys, aggregated):
            metrics[k.format(mode)](v, batch_size)



import pytest
import torch

from torchcross.data import TaskSource, Task, TaskTarget, TaskDescription
from torchcross.data.metadataset import (
    MetaDataset,
    MetaConcatDataset,
    IterableMetaDataset,
    MetaChainDataset,
)


class DummyTaskSource(TaskSource):
    data = torch.rand(10)
    labels = torch.randint(0, 2, (10,))
    task_description = TaskDescription(
        TaskTarget.MULTICLASS_CLASSIFICATION, {0: "a", 1: "b"}, task_identifier="test"
    )

    def __getitem__(self, index):
        return self.data[index], self.labels[index]


class DummyMetaDataset(MetaDataset):
    task_source = DummyTaskSource()

    def __getitem__(self, index):
        return Task(
            self.task_source[index],
            self.task_source[-index],
            self.task_source.task_description,
        )


class DummyIterableMetaDataset(IterableMetaDataset):
    task_source = DummyTaskSource()

    def __iter__(self):
        for i in range(10):
            yield Task(
                self.task_source[i],
                self.task_source[-i],
                self.task_source.task_description,
            )


class TestMetaDataset:
    def test_get_item(self):
        ds = DummyMetaDataset()
        task = ds[0]
        assert isinstance(task, Task)

    def test_addition(self):
        ds1 = DummyMetaDataset()
        ds2 = DummyMetaDataset()
        concat_ds = ds1 + ds2
        assert isinstance(concat_ds, MetaConcatDataset)


class TestMetaConcatDataset:
    def test_init(self):
        with pytest.raises(AssertionError):
            MetaConcatDataset([])

        ds1 = DummyMetaDataset()
        ds2 = DummyMetaDataset()
        concat_ds = MetaConcatDataset([ds1, ds2])
        assert len(concat_ds.datasets) == 2

    def test_get_item(self):
        ds1 = DummyMetaDataset()
        ds2 = DummyMetaDataset()
        concat_ds = MetaConcatDataset([ds1, ds2])
        task = concat_ds[(0, 0)]
        assert isinstance(task, Task)


class TestIterableMetaDataset:
    def test_iter(self):
        ds = DummyIterableMetaDataset()
        tasks = list(ds)
        assert len(tasks) == 10
        assert all(isinstance(task, Task) for task in tasks)

    def test_addition(self):
        ds1 = DummyIterableMetaDataset()
        ds2 = DummyIterableMetaDataset()
        chain_ds = ds1 + ds2
        assert isinstance(chain_ds, MetaChainDataset)


class TestMetaChainDataset:
    def test_init(self):
        ds1 = DummyIterableMetaDataset()
        ds2 = DummyIterableMetaDataset()
        chain_ds = MetaChainDataset([ds1, ds2])
        assert len(chain_ds.datasets) == 2



import random
from collections.abc import Iterable

from torch.utils.data import ChainDataset, IterableDataset


class InterleaveDataset(ChainDataset):

    datasets: Iterable[IterableDataset]

    def __init__(self, datasets: Iterable[IterableDataset]):
        super().__init__(datasets)

    def __iter__(self):
        iterators = [iter(dataset) for dataset in self.datasets]
        while iterators:
            to_remove = []
            for i in range(len(iterators)):
                try:
                    yield next(iterators[i])
                except StopIteration:
                    to_remove.append(i)
            for i in reversed(to_remove):
                del iterators[i]


class RandomInterleaveDataset(InterleaveDataset):
    def __init__(self, datasets: Iterable[IterableDataset]):
        super().__init__(datasets)

    def __iter__(self):
        iterators = [iter(dataset) for dataset in self.datasets]
        weights = [len(dataset) for dataset in self.datasets]
        while iterators:
            i = random.choices(range(len(iterators)), weights=weights)[0]
            weights[i] -= 1
            try:
                yield next(iterators[i])
            except StopIteration:
                del iterators[i]
                del weights[i]



from typing import overload, Iterable

from torch.utils.data import Dataset

from torchcross.data import TaskDescription, Task
from torchcross.data.metadataset import MetaDataset


class CollectionMetaDataset(MetaDataset):
    tasks: list[Task]

    @overload
    def __init__(self, tasks: Iterable[Task]) -> None:
        ...

    @overload
    def __init__(
        self,
        support_datasets: Iterable[Dataset],
        query_datasets: Iterable[Dataset],
        task_descriptions: Iterable[TaskDescription],
    ) -> None:
        ...

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        if len(args) == 1 and not kwargs:
            (tasks,) = args
            assert len(tasks) > 0
            self.tasks = list(tasks)
        elif len(args) == 3 and not kwargs:
            support_datasets, query_datasets, task_descriptions = args
            assert (
                len(support_datasets) == len(query_datasets) == len(task_descriptions)
            )
            assert len(support_datasets) > 0
            self.tasks = [
                Task(support, query, task_description)
                for support, query, task_description in zip(
                    support_datasets, query_datasets, task_descriptions
                )
            ]
        else:
            raise TypeError(f"Expected 1 or 3 positional arguments, got {len(args)}")

    def __getitem__(self, index: int) -> Task:
        return self.tasks[index]



from typing import Callable, Optional

import torch
import torch.nn.functional as F
from torch import nn

from torchcross.data.task import TaskTarget, TaskDescription

__all__ = ["get_loss_func"]


def get_loss_func(
    task_target: TaskTarget,
    classes: dict[int, str],
    device: torch.device = None,
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    """Get the loss function for a given task target and classes.

    Args:
        task_target: The task target.
        classes: The classes for the task.
        device: The device to create tensors needed for the loss function
            on.

    Returns:
        A loss function that takes logits and labels and returns a loss
        tensor.
    """
    match task_target:
        case TaskTarget.MULTICLASS_CLASSIFICATION:
            return lambda p, y, pos_class_weights=None: F.cross_entropy(
                p, y, weight=pos_class_weights
            )
        case TaskTarget.MULTILABEL_CLASSIFICATION | TaskTarget.BINARY_CLASSIFICATION:
            return (
                lambda p, y, pos_class_weights=None: F.binary_cross_entropy_with_logits(
                    p, y.float(), pos_weight=pos_class_weights
                )
            )
        case TaskTarget.ORDINAL_REGRESSION:
            class_to_idx = torch.full(
                (max(classes) + 1,),
                -1,
                dtype=torch.long,
                device=device,
            )
            for i, c in enumerate(classes):
                class_to_idx[c] = i
            return lambda p, y, pos_class_weights=None: F.cross_entropy(
                p, class_to_idx[y], weight=pos_class_weights
            )
        case TaskTarget.REGRESSION:
            return F.mse_loss
        case target:
            raise NotImplementedError(f"Unsupported task target: {target}")


class IndexedCrossEntropyLoss(nn.CrossEntropyLoss):
    def __init__(
        self,
        classes: dict[int, str],
        weight: Optional[torch.Tensor] = None,
        size_average=None,
        ignore_index: int = -100,
        reduce=None,
        reduction: str = "mean",
        label_smoothing: float = 0.0,
    ) -> None:
        super().__init__(
            weight, size_average, ignore_index, reduce, reduction, label_smoothing
        )

        self.register_buffer(
            "class_to_idx", torch.full((max(classes) + 1,), -1, dtype=torch.long)
        )
        for i, c in enumerate(classes):
            self.class_to_idx[c] = i

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        return super().forward(input, self.class_to_idx[target])


class FloatBCEWithLogitsLoss(nn.BCEWithLogitsLoss):
    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        return super().forward(input, target.float())


def get_criterion(
    task_description: TaskDescription,
    pos_class_weights: torch.Tensor = None,
    device: torch.device = None,
    **criterion_kwargs,
):
    match task_description.task_target:
        case TaskTarget.MULTICLASS_CLASSIFICATION:
            return nn.CrossEntropyLoss(weight=pos_class_weights, **criterion_kwargs)
        case TaskTarget.MULTILABEL_CLASSIFICATION | TaskTarget.BINARY_CLASSIFICATION:
            return FloatBCEWithLogitsLoss(
                pos_weight=pos_class_weights, **criterion_kwargs
            )
        case TaskTarget.ORDINAL_REGRESSION:
            return IndexedCrossEntropyLoss(
                task_description.classes, weight=pos_class_weights, **criterion_kwargs
            ).to(device)
        case TaskTarget.REGRESSION:
            return nn.MSELoss(**criterion_kwargs)
        case target:
            raise NotImplementedError(f"Unsupported task target: {target}")


def get_ordinal_cross_entropy_loss(classes):
    # Compute the weights for ordinal cross-entropy loss
    # using the intervals between consecutive class keys.
    sorted_class_keys = sorted(classes.keys())
    sorted_keys_tensor = torch.tensor(sorted_class_keys, dtype=torch.float)
    label_intervals_tensor = torch.diff(sorted_keys_tensor)
    weights = torch.cat((label_intervals_tensor, torch.tensor([1])), dim=0)

    def ordinal_cross_entropy(p, y):
        y = y.to(torch.long).unsqueeze(1)
        y_binary = y <= torch.arange(p.size(1), device=y.device)

        weightss = weights.unsqueeze(0).repeat(y.size(0), 1).to(p.device)

        binary_losses = F.binary_cross_entropy_with_logits(p, y_binary.float())
        weighted_binary_losses = binary_losses * weightss
        ordinal_loss = weighted_binary_losses.mean()
        return ordinal_loss

    return ordinal_cross_entropy



from functools import partial
from typing import Callable, Any, Optional, Literal

import torch
import torchmetrics
from torchmetrics import Metric
from torchmetrics.classification import (
    MulticlassAUROC,
    BinaryAUROC,
    MultilabelAUROC,
    MulticlassAccuracy,
    BinaryAccuracy,
    MultilabelAccuracy,
    Accuracy as _Accuracy,
    AUROC as _AUROC,
)

from torchcross.data.task import TaskTarget, TaskDescription


def get_accuracy_func(
    task_target: TaskTarget, classes: dict[int, str], device: torch.device = None
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    """Get the accuracy function for a given task target and classes.

    Args:
        task_target: The task target.
        classes: The classes for the task.
        device: The device to create tensors needed for the accuracy
            function on.

    Returns:
        A torchmetrics functional metric or a function wrapping a
        torchmetrics functional metric for the given task target and
        classes.
    """
    match task_target:
        case TaskTarget.MULTICLASS_CLASSIFICATION:
            return partial(
                torchmetrics.functional.classification.multiclass_accuracy,
                num_classes=len(classes),
                average="micro",
            )
        case TaskTarget.MULTILABEL_CLASSIFICATION:
            return partial(
                torchmetrics.functional.classification.multilabel_accuracy,
                num_labels=len(classes),
                average="micro",
            )
        case TaskTarget.BINARY_CLASSIFICATION:
            return torchmetrics.functional.classification.binary_accuracy
        case TaskTarget.ORDINAL_REGRESSION:
            class_to_idx = torch.full(
                (max(classes) + 1,),
                -1,
                dtype=torch.long,
                device=device,
            )
            for i, c in enumerate(classes):
                class_to_idx[c] = i
            acc = partial(
                torchmetrics.functional.classification.multiclass_accuracy,
                num_classes=len(classes),
                average="micro",
            )
            return lambda p, y: acc(p, class_to_idx[y])
        case target:
            raise NotImplementedError(f"Unsupported task target: {target}")


def get_auroc_func(
    task_target: TaskTarget, classes: dict[int, str], device: torch.device = None
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    """Get the area under the ROC curve metric function for a given task
    target and classes.

    Args:
        task_target: The task target.
        classes: The classes for the task.
        device: The device to create tensors needed for the metric
            function on.

    Returns:
        A torchmetrics functional metric or a function wrapping a
        torchmetrics functional metric for the given task target and
        classes.
    """
    match task_target:
        case TaskTarget.MULTICLASS_CLASSIFICATION:
            return partial(
                torchmetrics.functional.classification.multiclass_auroc,
                num_classes=len(classes),
            )
        case TaskTarget.MULTILABEL_CLASSIFICATION:
            return partial(
                torchmetrics.functional.classification.multilabel_auroc,
                num_labels=len(classes),
            )
        case TaskTarget.BINARY_CLASSIFICATION:
            return torchmetrics.functional.classification.binary_auroc
        case TaskTarget.ORDINAL_REGRESSION:
            class_to_idx = torch.full(
                (max(classes) + 1,),
                -1,
                dtype=torch.long,
                device=device,
            )
            for i, c in enumerate(classes):
                class_to_idx[c] = i
            auroc = partial(
                torchmetrics.functional.classification.multiclass_auroc,
                num_classes=len(classes),
            )
            return lambda p, y: auroc(p, class_to_idx[y])
        case target:
            raise NotImplementedError(f"Unsupported task target: {target}")


class IndexedMulticlassAccuracy(MulticlassAccuracy):
    def __init__(
        self,
        classes: dict[int, str],
        top_k: Optional[int] = 1,
        average: Optional[Literal["micro", "macro", "weighted", "none"]] = "micro",
        multidim_average: Literal["global", "samplewise"] = "global",
        ignore_index: Optional[int] = None,
        validate_args: bool = True,
    ) -> None:
        super().__init__(
            len(classes),
            top_k=top_k,
            average=average,
            multidim_average=multidim_average,
            ignore_index=ignore_index,
            validate_args=validate_args,
        )
        self.register_buffer("class_to_idx", torch.full(
            (max(classes) + 1,), -1, dtype=torch.long, device=self.device
        ))
        for i, c in enumerate(classes):
            self.class_to_idx[c] = i

    def update(self, preds: torch.Tensor, target: torch.Tensor) -> None:
        super().update(preds, self.class_to_idx[target])


class Accuracy(_Accuracy):
    def __new__(
        cls,
        task_description: TaskDescription,
        *,
        top_k: int = 1,
        threshold: float = 0.5,
        average: Optional[Literal["micro", "macro", "weighted", "none"]] = "micro",
        multidim_average: Literal["global", "samplewise"] = "global",
        ignore_index: Optional[int] = None,
        validate_args: bool = True,
        **kwargs: Any,
    ) -> Metric:
        kwargs.update(
            dict(
                multidim_average=multidim_average,
                ignore_index=ignore_index,
                validate_args=validate_args,
            )
        )

        task_target = task_description.task_target
        classes = task_description.classes

        if task_target == TaskTarget.MULTICLASS_CLASSIFICATION:
            assert isinstance(classes, dict)
            assert isinstance(top_k, int)
            return MulticlassAccuracy(len(classes), top_k, average, **kwargs)
        if task_target == TaskTarget.MULTILABEL_CLASSIFICATION:
            assert isinstance(classes, dict)
            return MultilabelAccuracy(len(classes), threshold, average, **kwargs)
        if task_target == TaskTarget.BINARY_CLASSIFICATION:
            return BinaryAccuracy(threshold, **kwargs)
        if task_target == TaskTarget.ORDINAL_REGRESSION:
            assert isinstance(classes, dict)
            return IndexedMulticlassAccuracy(classes, top_k, average, **kwargs)
        if task_target == TaskTarget.REGRESSION:
            raise ValueError("Accuracy metric not defined for regression tasks")
        raise NotImplementedError(f"Unsupported task target: {task_target}")


class IndexedMulticlassAUROC(MulticlassAUROC):
    def __init__(
        self,
        classes: dict[int, str],
        average: Optional[Literal["macro", "weighted", "none"]] = "macro",
        thresholds: Optional[int | list[float] | torch.Tensor] = None,
        ignore_index: Optional[int] = None,
        validate_args: bool = True,
    ) -> None:
        super().__init__(
            len(classes),
            average=average,
            thresholds=thresholds,
            ignore_index=ignore_index,
            validate_args=validate_args,
        )
        self.register_buffer("class_to_idx", torch.full(
            (max(classes) + 1,), -1, dtype=torch.long, device=self.device
        ))
        for i, c in enumerate(classes):
            self.class_to_idx[c] = i

    def update(self, preds: torch.Tensor, target: torch.Tensor) -> None:
        super().update(preds, self.class_to_idx[target])


class AUROC(_AUROC):
    def __new__(
        cls,
        task_description: TaskDescription,
        *,
        average: Optional[Literal["macro", "weighted", "none"]] = "macro",
        max_fpr: Optional[float] = None,
        thresholds: Optional[int | list[float] | torch.Tensor] = None,
        ignore_index: Optional[int] = None,
        validate_args: bool = True,
        **kwargs: Any,
    ) -> Metric:
        kwargs.update(
            dict(
                thresholds=thresholds,
                ignore_index=ignore_index,
                validate_args=validate_args,
            )
        )

        task_target = task_description.task_target
        classes = task_description.classes

        if task_target == TaskTarget.MULTICLASS_CLASSIFICATION:
            assert isinstance(classes, dict)
            return MulticlassAUROC(len(classes), average, **kwargs)
        if task_target == TaskTarget.MULTILABEL_CLASSIFICATION:
            assert isinstance(classes, dict)
            return MultilabelAUROC(len(classes), average, **kwargs)
        if task_target == TaskTarget.BINARY_CLASSIFICATION:
            return BinaryAUROC(max_fpr, **kwargs)
        if task_target == TaskTarget.ORDINAL_REGRESSION:
            assert isinstance(classes, dict)
            return IndexedMulticlassAUROC(classes, average, **kwargs)
        if task_target == TaskTarget.REGRESSION:
            raise ValueError("AUROC metric not defined for regression tasks")
        raise NotImplementedError(f"Unsupported task target: {task_target}")



import logging
from typing import Callable, Any, Literal

import torch
from torch import nn
from torch.optim import Optimizer
from torchmetrics import MetricCollection

from torchcross import models
from torchcross.cd.metrics import Accuracy, AUROC
from torchcross.cd.metrics.wrappers import CrossDomainMeanWrapper
from torchcross.data.task import TaskDescription
from torchcross.models.lightning.cross_domain import CrossDomainLightningModule


class SimpleCrossDomainClassifier(
    models.SimpleCrossDomainClassifier, CrossDomainLightningModule
):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        optimizer: type[Optimizer],
        lr_scheduler: Callable[[Optimizer], Any] | None = None,
        task_descriptions: list[TaskDescription] = None,
        add_heads_during_training: bool = True,
        scale_loss_by_batch_count: bool = False,
    ):
        super().__init__(
            backbone,
            num_backbone_features,
            task_descriptions,
            add_heads_during_training,
        )

        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

        self.scale_loss_by_batch_count = scale_loss_by_batch_count

        self.training_metrics = MetricCollection(
            {
                "accuracy": CrossDomainMeanWrapper(Accuracy, task_descriptions),
                "AUROC": CrossDomainMeanWrapper(AUROC, task_descriptions),
            },
            postfix="/train",
        )
        self.validation_metrics = self.training_metrics.clone(postfix="/val")
        self.test_metrics = self.training_metrics.clone(postfix="/test")

    def add_head(self, task_description: TaskDescription) -> nn.Module:
        new_head = super().add_head(task_description)
        self.optimizers().add_param_group({"params": new_head.parameters()})
        return new_head

    def call_metrics(
        self,
        mode: Literal["train", "val", "test"],
        task_description: TaskDescription,
        pred: torch.Tensor,
        y: torch.Tensor,
    ):
        if mode == "train":
            metrics = self.training_metrics
        elif mode == "val":
            metrics = self.validation_metrics
        elif mode == "test":
            metrics = self.test_metrics
        else:
            raise ValueError(f"Invalid mode {mode}")
        return metrics(task_description, pred, y)

    def validation_step(
        self,
        batch: tuple[tuple[torch.Tensor, torch.Tensor], TaskDescription],
        batch_idx: int,
    ):
        try:
            super().validation_step(batch, batch_idx)
        except ValueError as e:
            if "Head for task" in str(e):
                logging.warning(f"{e} Skipping validation step.")
            else:
                raise e



import bisect
from abc import ABC
from collections.abc import Callable, Sequence

import numpy as np
import torch
from torch.utils.data import (
    Dataset,
    ConcatDataset,
    BatchSampler,
    RandomSampler,
    SequentialSampler,
    IterableDataset,
)

from torchcross.utils.collate_fn import identity
from .task import TaskTarget, TaskDescription

__all__ = [
    "TaskSource",
    "WrapTaskSource",
    "BatchedTaskSource",
    "ConcatTaskSource",
]

from ..utils import to_numpy


class TaskSource(Dataset, ABC):
    """A dataset that can be used as a source to generate tasks."""

    task_description: TaskDescription
    labels: np.ndarray | torch.Tensor | Sequence

    def get_num_samples_per_class(
        self, neg=False
    ) -> np.ndarray | torch.Tensor | list[int]:
        task_target = self.task_description.task_target
        total = len(self.labels)

        classes = self.task_description.classes
        num_classes = len(classes)
        max_class_key = max(classes.keys())

        not_implemented_message = f"Task target {task_target} not yet implemented"

        if isinstance(self.labels, np.ndarray):
            if task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
                num_samples = np.bincount(self.labels, minlength=num_classes)
            elif task_target is TaskTarget.MULTILABEL_CLASSIFICATION:
                num_samples = np.sum(self.labels, axis=0)
            elif task_target is TaskTarget.BINARY_CLASSIFICATION:
                num_samples = np.sum(self.labels, axis=0)
            elif task_target is TaskTarget.ORDINAL_REGRESSION:
                num_samples = np.bincount(self.labels, minlength=max_class_key + 1)
                num_samples = num_samples[np.array(list(classes.keys()), dtype=np.int_)]
            else:
                raise NotImplementedError(not_implemented_message)
            return num_samples - total if neg else num_samples

        elif isinstance(self.labels, torch.Tensor):
            if task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
                num_samples = torch.bincount(self.labels, minlength=num_classes)
            elif task_target is TaskTarget.MULTILABEL_CLASSIFICATION:
                num_samples = torch.sum(self.labels, dim=0)
            elif task_target is TaskTarget.BINARY_CLASSIFICATION:
                num_samples = torch.sum(self.labels, dim=0)
            elif task_target is TaskTarget.ORDINAL_REGRESSION:
                num_samples = torch.bincount(self.labels, minlength=max_class_key + 1)
                num_samples = num_samples[
                    torch.tensor(
                        list(classes.keys()),
                        device=num_samples.device,
                        dtype=torch.long,
                    )
                ]
            else:
                raise NotImplementedError(not_implemented_message)
            return num_samples - total if neg else num_samples

        elif isinstance(self.labels, Sequence):
            if task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
                num_samples = [sum(l == c for l in self.labels) for c in classes]
            elif task_target is TaskTarget.MULTILABEL_CLASSIFICATION:
                num_samples = [sum(l[c] for l in self.labels) for c in classes]
            elif task_target is TaskTarget.BINARY_CLASSIFICATION:
                num_samples = [
                    sum(l[0] if isinstance(l, Sequence) else l for l in self.labels)
                ]
            elif task_target is TaskTarget.ORDINAL_REGRESSION:
                num_samples = [sum(l == c for l in self.labels) for c in classes]
            else:
                raise NotImplementedError(not_implemented_message)
            return [n - total for n in num_samples] if neg else num_samples

        else:
            raise ValueError("Unsupported label data type")


class WrapTaskSource(TaskSource):
    """Wraps a dataset as a task source."""

    def __init__(
        self,
        dataset: Dataset,
        task_target: TaskTarget,
        classes: dict[int, str],
        label_fn: Callable[[Dataset], np.ndarray | torch.Tensor | Sequence] = None,
        task_identifier: str = "",
        domain_identifier: str = "",
    ):
        self.dataset = dataset
        self.task_description = TaskDescription(
            task_target, classes, task_identifier, domain_identifier
        )

        def default_label_fn(ds: Dataset):
            return np.stack([to_numpy(item[1]) for item in ds])

        label_fn = label_fn if label_fn else default_label_fn
        self.labels = label_fn(self.dataset)

    def __getitem__(self, index):
        return self.dataset[index]

    def __len__(self):
        return len(self.dataset)


class BatchedTaskSource(IterableDataset):
    def __init__(
        self,
        task_source: TaskSource,
        batch_size: int,
        shuffle: bool = True,
        drop_last: bool = False,
        collate_fn=None,
        with_task_description: bool = False,
    ):
        self.task_source = task_source
        self.task_description = task_source.task_description
        self.with_task_description = with_task_description

        sampler = (
            RandomSampler(self.task_source)
            if shuffle
            else SequentialSampler(self.task_source)
        )
        self.batch_sampler = BatchSampler(sampler, batch_size, drop_last)
        self.collate_fn = collate_fn if collate_fn else identity

    def __iter__(self):
        for batch in self.batch_sampler:
            if self.with_task_description:
                yield (
                    self.collate_fn([self.task_source[i] for i in batch]),
                    self.task_description,
                )
            else:
                yield self.collate_fn([self.task_source[i] for i in batch])

    def __len__(self):
        return len(self.batch_sampler)


class ConcatTaskSource(TaskSource, ConcatDataset):
    datasets: list[TaskSource]

    def __getitem__(self, idx):
        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    "absolute value of index should not exceed dataset length"
                )
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return (
            self.datasets[dataset_idx][sample_idx],
            self.datasets[dataset_idx].task_description,
        )

    def __len__(self):
        return sum(len(d) for d in self.datasets)


class SubsetTaskSource(TaskSource):
    def __init__(self, task_source: TaskSource, indices: Sequence[int]):
        self.task_source = task_source
        self.indices = indices
        self.task_description = task_source.task_description
        self.labels = task_source.labels[indices]

    def __getitem__(self, idx):
        return self.task_source[self.indices[idx]]

    def __len__(self):
        return len(self.indices)



from collections.abc import Collection
from dataclasses import dataclass
from enum import Enum, auto

from torch.utils.data import Dataset


class TaskTarget(Enum):
    MULTICLASS_CLASSIFICATION = auto()
    BINARY_CLASSIFICATION = auto()
    MULTILABEL_CLASSIFICATION = auto()
    ORDINAL_REGRESSION = auto()
    REGRESSION = auto()
    BINARY_SEGMENTATION = auto()
    MULTILABEL_SEGMENTATION = auto()


@dataclass
class TaskDescription:
    """A task description with a task target and classes as well as
    optional string identifiers for the task and domain.

    Args:
        task_target: The task target.
        classes: The classes for the task.
        task_identifier: The task identifier.
        domain_identifier: The domain identifier.
    """

    task_target: TaskTarget
    classes: dict[int, str] = None
    task_identifier: str = None
    domain_identifier: str = None


@dataclass
class Task:
    """A task with support and query datasets as well as a task
    description.

    Args:
        support: The support dataset or collection of inputs and targets
        query: The query dataset or collection of inputs and targets
        description: The task description
    """

    support: Dataset | Collection
    query: Dataset | Collection
    description: TaskDescription



from collections import OrderedDict
from typing import Literal

import torch
from torch import nn, Tensor

from torchcross.cd.activations import get_log_prob_func
from torchcross.data.task import Task, TaskTarget
from torchcross.models import CrossDomainModel
from torchcross.models.episodic import EpisodicModel
from torchcross.utils.layers import Expand


class ProtoNet(EpisodicModel):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        distance: Literal[
            "euclidean", "cosine", "squared_euclidean"
        ] = "squared_euclidean",
    ) -> None:
        super(ProtoNet, self).__init__()

        self.backbone = backbone
        self.num_backbone_features = num_backbone_features

        self.distance = distance

    def compute_distance(self, x: Tensor, y: Tensor) -> Tensor:
        if self.distance == "euclidean":
            return torch.cdist(x, y)
        elif self.distance == "cosine":
            return 1 - torch.cosine_similarity(x, y)
        elif self.distance == "squared_euclidean":
            return torch.pow(x[:, None] - y[None, :], 2).sum(dim=2)
        else:
            raise ValueError(f"Distance {self.distance} not implemented")

    def episode(self, task: Task) -> torch.Tensor:
        prototypes, neg_prototypes = self.get_prototypes(task)
        query_x = task.query[0]
        query_features = self.backbone(query_x)
        dist = self.compute_distance(query_features, prototypes)
        logits_func = get_log_prob_func(task.description.task_target)
        query_logits = logits_func(-dist)
        if neg_prototypes is not None:
            neg_dist = self.compute_distance(query_features, neg_prototypes)
            neg_logits = logits_func(-neg_dist)
            query_logits = query_logits - neg_logits
        return query_logits

    def get_prototypes(self, task: Task) -> tuple[Tensor, Tensor]:
        support_x, support_y = task.support
        support_features = self.backbone(support_x)
        task_target = task.description.task_target
        classes = task.description.classes

        if task_target is TaskTarget.MULTICLASS_CLASSIFICATION:
            prototypes = torch.stack(
                [support_features[support_y == c].mean(dim=0) for c in classes]
            )
            neg_prototypes = None
        elif task_target is TaskTarget.MULTILABEL_CLASSIFICATION:
            prototypes = torch.stack(
                [support_features[support_y[:, c] == 1].mean(dim=0) for c in classes]
            )
            neg_prototypes = torch.stack(
                [support_features[support_y[:, c] == 0].mean(dim=0) for c in classes]
            )
        elif task_target is TaskTarget.BINARY_CLASSIFICATION:
            prototypes = torch.stack([support_features[support_y == 1].mean(dim=0)])
            neg_prototypes = torch.stack([support_features[support_y == 0].mean(dim=0)])
        elif task_target is TaskTarget.ORDINAL_REGRESSION:
            # treat as multiclass classification for now
            prototypes = torch.stack(
                [support_features[support_y == c].mean(dim=0) for c in classes]
            )
            neg_prototypes = None
        else:
            raise NotImplementedError(f"Task target {task_target} not yet implemented")
        return prototypes, neg_prototypes


class SimpleCrossDomainProtoNet(ProtoNet, CrossDomainModel):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        distance: Literal[
            "euclidean", "cosine", "squared_euclidean"
        ] = "squared_euclidean",
    ) -> None:
        super().__init__(backbone, num_backbone_features, distance)

        self.backbone = torch.nn.Sequential(
            OrderedDict(input_module=Expand(-1, 3, -1, -1), backbone=self.backbone)
        )



from collections.abc import Callable, Sequence, Iterable
from typing import Optional, Any

import torch
import torchopt
from torch import Tensor, nn
from torch.optim import Optimizer

from .episodic import EpisodicLightningModule
from ... import models
from ...cd.activations import get_prob_func
from ...cd.losses import get_loss_func
from ...cd.metrics import get_accuracy_func, get_auroc_func
from ...data import TaskDescription, Task


class MAML(models.MAML, EpisodicLightningModule):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        task_description: TaskDescription,
        outer_optimizer: Callable[
            [Iterable[Tensor] | Iterable[dict[str, Any]]], torch.optim.Optimizer
        ],
        inner_optimizer: Callable[[torch.nn.Module], torchopt.MetaOptimizer],
        eval_inner_optimizer: Callable[
            [Sequence[torch.nn.Parameter]], torchopt.Optimizer
        ],
        num_inner_steps: int,
        eval_num_inner_steps: Optional[int] = None,
        transductive: bool = True,
        reset_head: Optional[models.maml.head_init_t] = None,
        outer_lr_scheduler: Callable[[Optimizer], Any] | None = None,
    ) -> None:
        super(MAML, self).__init__(
            backbone,
            num_backbone_features,
            task_description,
            inner_optimizer,
            eval_inner_optimizer,
            num_inner_steps,
            eval_num_inner_steps,
            transductive,
            reset_head,
        )

        self.outer_optimizer = outer_optimizer
        self.outer_lr_scheduler = outer_lr_scheduler
        metric_keys = [
            "accuracy/meta{}/support",
            "accuracy/meta{}/query",
            "AUROC/meta{}/support",
            "AUROC/meta{}/query",
        ]
        self.configure_metrics(metric_keys)

        self.accuracy_func = get_accuracy_func(
            task_description.task_target, task_description.classes, self.device
        )
        self.auroc_func = get_auroc_func(
            task_description.task_target, task_description.classes, self.device
        )

    def compute_metrics(self, task, support_pred, query_pred):
        support_y = task.support[1]
        query_y = task.query[1]
        support_accuracy = self.accuracy_func(support_pred, support_y)
        query_accuracy = self.accuracy_func(query_pred, query_y)
        support_auroc = self.auroc_func(support_pred, support_y)
        query_auroc = self.auroc_func(query_pred, query_y)
        return support_accuracy, query_accuracy, support_auroc, query_auroc

    def get_metrics_and_losses(
        self, batch: Sequence[Task]
    ) -> tuple[list[tuple[Tensor, ...]], list[Tensor], list[Tensor]]:
        net_state_dict = torchopt.extract_state_dict(
            self.net, by="reference", detach_buffers=True
        )
        metric_values = []
        support_losses = []
        query_losses = []
        for task in batch:
            support_x, support_y = task.support
            query_x, query_y = task.query

            loss_func = get_loss_func(
                task.description.task_target, task.description.classes, self.device
            )
            pred_func = get_prob_func(task.description.task_target)

            support_logit, query_logit = self.episode(task, with_support_output=True)
            query_losses.append(loss_func(query_logit, query_y))
            with torch.no_grad():
                support_losses.append(loss_func(support_logit, support_y))
                support_pred = pred_func(support_logit)
                query_pred = pred_func(query_logit)
                metric_values.append(
                    self.compute_metrics(task, support_pred, query_pred)
                )

            torchopt.recover_state_dict(self.net, net_state_dict)
        return metric_values, support_losses, query_losses


class CrossDomainMAML(models.CrossDomainMAML, EpisodicLightningModule):
    def __init__(
        self,
        backbone: nn.Module,
        num_backbone_features: int,
        outer_optimizer: Callable[
            [Iterable[Tensor] | Iterable[dict[str, Any]]], torch.optim.Optimizer
        ],
        inner_optimizer: Callable[[torch.nn.Module], torchopt.MetaOptimizer],
        eval_inner_optimizer: Callable[
            [Sequence[torch.nn.Parameter]], torchopt.Optimizer
        ],
        num_inner_steps: int,
        eval_num_inner_steps: Optional[int] = None,
        transductive: bool = True,
        head_init: Optional[models.maml.head_init_t] = None,
        outer_lr_scheduler: Callable[[Optimizer], Any] | None = None,
    ) -> None:
        super(CrossDomainMAML, self).__init__(
            backbone,
            num_backbone_features,
            inner_optimizer,
            eval_inner_optimizer,
            num_inner_steps,
            eval_num_inner_steps,
            transductive,
            head_init,
        )

        self.outer_optimizer = outer_optimizer
        self.outer_lr_scheduler = outer_lr_scheduler
        metric_keys = [
            "accuracy/meta{}/support",
            "accuracy/meta{}/query",
            "AUROC/meta{}/support",
            "AUROC/meta{}/query",
        ]
        self.configure_metrics(metric_keys)

    def compute_metrics(self, task, support_pred, query_pred):
        accuracy_func = get_accuracy_func(
            task.description.task_target, task.description.classes, self.device
        )
        auroc_func = get_auroc_func(
            task.description.task_target, task.description.classes, self.device
        )
        support_y = task.support[1]
        query_y = task.query[1]
        support_accuracy = accuracy_func(support_pred, support_y)
        query_accuracy = accuracy_func(query_pred, query_y)
        support_auroc = auroc_func(support_pred, support_y)
        query_auroc = auroc_func(query_pred, query_y)
        return support_accuracy, query_accuracy, support_auroc, query_auroc

    def get_metrics_and_losses(
        self, batch: Sequence[Task]
    ) -> tuple[list[tuple[Tensor, ...]], list[Tensor], list[Tensor]]:
        backbone_state_dict = torchopt.extract_state_dict(
            self.backbone, by="reference", detach_buffers=True
        )
        metric_values = []
        support_losses = []
        query_losses = []
        for task in batch:
            support_x, support_y = task.support
            query_x, query_y = task.query

            loss_func = get_loss_func(
                task.description.task_target, task.description.classes, self.device
            )
            pred_func = get_prob_func(task.description.task_target)

            support_logit, query_logit = self.episode(task, with_support_output=True)
            query_losses.append(loss_func(query_logit, query_y))
            with torch.no_grad():
                support_losses.append(loss_func(support_logit, support_y))
                support_pred = pred_func(support_logit)
                query_pred = pred_func(query_logit)
                metric_values.append(
                    self.compute_metrics(task, support_pred, query_pred)
                )

            torchopt.recover_state_dict(self.backbone, backbone_state_dict)
        return metric_values, support_losses, query_losses



from collections.abc import Sequence

import numpy as np
import torch


def to_numpy(arg, /):
    """Convert torch tensors, lists, tuples, other Sequences and sets to
    numpy arrays. Numpy arrays are returned as is. Single numbers are
    converted to numpy arrays of size 1. Collections are converted
    recursively.
    """

    if isinstance(arg, np.ndarray):
        return arg  # Already a numpy array, no conversion necessary
    elif isinstance(arg, torch.Tensor):
        return arg.detach().cpu().numpy()  # Make sure to detach and move to CPU
    elif isinstance(arg, (list, tuple, Sequence)):
        return np.array([to_numpy(item) for item in arg])
    elif isinstance(arg, set):
        return np.array([to_numpy(item) for item in sorted(arg)])
    # think about how to convert dict and other mappings
    # elif isinstance(collection, dict):
    #     return {key: to_numpy(value) for key, value in collection.items()}
    else:
        # This will work for single numbers, etc., but may fail with other types
        return np.array(arg)
